#include <assert.h>
#include <string>
#include <fstream>
#include <vector>
#include <map>
#include <algorithm>
#include <iostream>
#include <sstream> 
#include "vector_inl.h"
#include "GLee.h"
#include <GL/glut.h>
#include <GL/Soil.h>

#ifndef _MESH_H_
#define _MESH_H_

namespace MeshTools {

using namespace std;

const vector<string> defaultGroups;
struct TexCoord {
  Scalar u,v;

  inline TexCoord() : u(0),v(0) {};
  inline TexCoord(double u,double v) : u(u),v(v) {};
};

struct Face {
  size_t      numVertices;
  vector<int> vertexIdx;
  vector<int> normalIdx;
  vector<int> flatNormalIdx;
  vector<int> smoothNormalIdx;
  vector<int> textureIdx;

  Face() : numVertices(0) { };
  void Clear() { 
    numVertices= 0;
    vertexIdx.clear();
    flatNormalIdx.clear();
    textureIdx.clear();
  };
};

struct rgba {
  union {
    struct {float r,g,b,a;};
    float arr_rgba[4];
  };

  rgba::rgba() : r(0), g(0), b(0), a(0) {};
};

/* here here here
map_Ka -s 1 1 1 -o 0 0 0 -mm 0 1 chrome.mpc
map_Kd -s 1 1 1 -o 0 0 0 -mm 0 1 chrome.mpc
map_Ks -s 1 1 1 -o 0 0 0 -mm 0 1 chrome.mpc
map_Ns -s 1 1 1 -o 0 0 0 -mm 0 1 wisp.mps
map_d  -s 1 1 1 -o 0 0 0 -mm 0 1 wisp.mps
disp   -s 1 1 .5                 wisp.mps
decal  -s 1 1 1 -o 0 0 0 -mm 0 1 sand.mps
bump   -s 1 1 1 -o 0 0 0 -bm 1   sand.mpb
*/

struct Material {
  float shininess;
  float emission;
  rgba diffuse;
  rgba ambient;
  rgba specular;
  string map_ka;
  string map_kd;

  Material::Material() : shininess(0), emission(0) {
    diffuse.r  = 0.2f; diffuse.g  = 0.2f; diffuse.b  = 0.2f; diffuse.a  = 1.0f;
    ambient.r  = 0.8f; ambient.g  = 0.8f; ambient.b  = 0.8f; ambient.a  = 1.0f;
    specular.r = 1.0f; specular.g = 1.0f; specular.b = 1.0f; specular.a = 1.0f;
   map_ka = "";
   map_kd = "";
  };

  float const GetShininess() {return shininess;};
  float* const GetDiffuse() {return diffuse.arr_rgba;};
  float* const GetAmbient() {return ambient.arr_rgba;};
  float* const GetSpecular() {return specular.arr_rgba;};
};

// draw what
// style
enum ShadeStyle{
	// shading modes - mutually exclusive
  ShadeStyleSTART=-1,
  objOriginal    ,
  flat           ,
  smooth         ,
  ShadeStyleEND
};
inline ShadeStyle &operator++(ShadeStyle &d)
{
  d = ShadeStyle(d + 1);
  d == ShadeStyleEND?d = objOriginal:d;
  return d;
}
inline ShadeStyle &operator--(ShadeStyle &d)
{
  d = ShadeStyle(d - 1);
  d == ShadeStyleSTART?d = smooth:d;
  return d;
}

enum DrawStyle{
  // what to draw - can be mixed
	points        = 4,
  lines         = 8,
  faces         = 16,
  objNormal     = 32,
  flatNormal    = 64,
  smoothNormal  = 128,
};

inline DrawStyle operator^ (const DrawStyle& ds1, const DrawStyle& ds2)
{
	return static_cast<DrawStyle>(static_cast<int>(ds1) ^ static_cast<int>(ds2));
}

struct Group {
  //vector<int> materials;
  string material;
  vector<int> faces;
  vector<int> tris;
  vector<int> quads;
  vector<int> polys;
};

struct Texture {
  string name;
  string fullpath;
  int    id;
};

class Mesh {
private:
// data
  size_t numVertices;
  size_t numNormals;
  size_t numTexCoords;
  size_t numFaces;
  size_t numTris;
  size_t numQuads;
  size_t numPolys;

  map<string, Group>    groups;        // a map of groups
  map<string, Material> materials;
  vector<Face>          faces;                // a list of faces (a face is a list of indices to vertices that make up a face)
  vector<Face>          tris;
  vector<Face>          quads;
  vector<Face>          polys;
  map<string, Texture>  textures;        // textures (bmp, png, etc) and full pathname

  vector<Vector3d>  vertices;
  vector<Vector3d>  normals;
  vector<TexCoord>  texCoords;
  vector<Vector3d>  smoothNormals;
  vector<Vector3d>  flatNormals;

  Vector3d upperBound;
  Vector3d lowerBound;
  Vector3d center;

  void SetFlatNormals(vector<Face> &faces, vector<Vector3d>& flatNormals);

  GLuint texMesh;


public:
  DrawStyle drawStyle;
  ShadeStyle shadeStyle;

  Mesh();
  //~Mesh();

// methods
  //accessors
  map<string, Texture> Textures() const {return textures;};
  map<string, Group> const Groups() const {return groups;};
  map<string, Material> const Materials() const {return materials;};
  Material& GetMaterial(string matName) {
	  return materials[matName];
  };

  size_t const NumFaces() const {return numFaces;};
  size_t const NumTris() const {return tris.size();};
  size_t const NumPolys() const {return polys.size();};
  size_t const NumQuads() const {return quads.size();};
  size_t const NumVertices() const {return numVertices;};
  Vector3d const GetVertex(int i) const {return vertices[i];};

  size_t const NumNormals() const {return normals.size();};
  Vector3d& Normal(int i) {return normals[i];};

  size_t const NumFlatNormals() const {return flatNormals.size();};
  Vector3d& FlatNormal(int i) {return flatNormals[i];};

  size_t const NumSmoothNormals() const {return smoothNormals.size();};
  Vector3d& SmoothNormal(int i) {return smoothNormals[i];};

  size_t const NumTexCoords() const {return texCoords.size();};

  size_t const NumObjNormals() const {return normals.size();};
  Vector3d& ObjNormal(int i) {return normals[i];};

  Scalar const GetWidth() const {return upperBound.y - lowerBound.y;};
  Vector3d const GetUpperBound() const {return upperBound;};
  Vector3d const GetLowerBound() const {return lowerBound;};

  void Center(Vector3d pos) {center = pos;};
  Vector3d const GetCenter() const {return center;};

  Face& GetFace(int i) {return faces[i];};
  Face& Tri(int i) {return tris[i];};
  Face& Quad(int i) {return quads[i];};
  Face& Poly(int i) {return polys[i];};

  void Load(string sFileName);

  // builder
  void AddFace(Face& face, const vector<string>& groups = defaultGroups);

  void AddVertex(const Vector3d& vertex);
  void AddNormal(const Vector3d& normal);
  void AddTexture(const TexCoord& texture);

  void AddVertexIndex(const int& idx);
  void AddNormalIndex(const int& idx);
  void AddTextureIndex(const int& idx);

  void AddVertexIndex(const int &idx, const int& faceIdx);
  void AddNormalIndex(const int &idx, const int& faceIdx);
  void AddTextureIndex(const int &idx, const int& faceIdx);

  void AddMaterial(string name, Material material);
  void SetMaterial(vector<string> group, string material);

  void AddTextureName(string name, string pathname);

  // utility
  void Clear();
  void SetNormals();
  void SetFlatNormals();
  void SetSmoothNormals();
  void FindBoundingBox();
  void InvertNormals();
  void Centralise();
  void ScaleBy(const float factor);

  void Render();
};

//////////////////////////////////////////////////////////////////////
// non-member functions - see Meyer
void AddVertices(Mesh mesh, const vector<Vector3d> vertices);
//void AddVertices(Mesh& mesh, const vector<string> vertices);

void AddNormals(Mesh mesh, const vector<Vector3d> normals);
//void AddNormals(Mesh mesh, const vector<string> normals);

void AddTextures(Mesh mesh, const vector<TexCoord> textures);
//void AddTextures(Mesh mesh, const vector<string> textures);

void AddVertexIndices(Mesh mesh, const vector<int> vertexIndicesFace);
//void AddVertexIndices(Mesh& mesh, const vector<string> vertexIndices);

void AddNormalIndices(Mesh mesh, const vector<int> normalIndices);
//void AddNormalIndices(Mesh& mesh, const vector<string> normalIndices);

void AddTextureIndices(Mesh mesh, const vector<int> textureIndices);
//void AddTextureIndices(Mesh& mesh, const vector<string> textureIndices);


void AddVertexIndex(Mesh mesh, const string idx, const int faceIdx);
void AddNormalIndex(Mesh mesh, const string idx, const int faceIdx);
void AddTextureIndex(Mesh mesh, const string idx, const int faceIdx);
}
#endif