#include "MeshDataLoader.h"
#include <gl/glut.h>
#include "soil.h"

Mesh* MeshDataLoader::mesh = NULL;
Face* MeshDataLoader::face = new Face;
int MeshDataLoader::currentFace = -1;
vector<string> MeshDataLoader::currentGroups;

std::map<FunctionIDs, PtrToFunVecStr> MeshDataLoader::AvailableFunctions() {
// returns a std::map of pointers to implemented functions and their associated function id

  std::map<FunctionIDs, PtrToFunVecStr> pfParseFuncs;

  pfParseFuncs[addVertex]   = MeshDataLoader::AddVertex;
  pfParseFuncs[addNormal]   = MeshDataLoader::AddNormal;
  pfParseFuncs[addTexture]  = MeshDataLoader::AddTexture;
  pfParseFuncs[addFaceV]    = MeshDataLoader::AddFaceV;
  pfParseFuncs[addFaceVT]   = MeshDataLoader::AddFaceVT;
  pfParseFuncs[addFaceVN]   = MeshDataLoader::AddFaceVN;
  pfParseFuncs[addFaceVTN]  = MeshDataLoader::AddFaceVTN;
  pfParseFuncs[setGroups]   = MeshDataLoader::SetGroups;
  pfParseFuncs[setMaterial] = MeshDataLoader::SetMaterial;
  pfParseFuncs[addMaterial] = MeshDataLoader::AddMaterial;

  return pfParseFuncs;
}

void MeshDataLoader::Load(const string filename, Mesh &mesh) {
  if (ObjFileReader::CanRead(filename)) {
    if (ObjFileReader::FnPtrsAreSet() == false) {
      ObjFileReader::SetExternalFunctionPointers(MeshDataLoader::AvailableFunctions());
    }
    mesh.Clear();
    currentFace = -1;

    MeshDataLoader::mesh = &mesh;
    ObjFileReader::Load(filename);
  }

  size_t nameStart = filename.find_last_of("\\");
  if(nameStart == string::npos)
    nameStart = filename.find_last_of("/");
  string path = filename.substr(0, nameStart+1);
  string modelName = filename.substr(nameStart+1);

  //return;
  // load any required textures
  if(mesh.Textures().size()) {
    std::cout << " textures" << endl;
    map<string, Texture>& meshTextures(mesh.Textures());
    for (map<string, Texture>::iterator i = meshTextures.begin(); i != meshTextures.end(); ++i)  {

      //string picfile(i->second.fullpath + i->second.name);
	  string picfile(path + i->second.name);

      /* load an image file directly as a new OpenGL texture */
      //GLuint tex_2d = SOIL_load_OGL_texture
      int tex_2d = SOIL_load_OGL_texture
      (
        picfile.c_str(),//"img_test.png", // 
        SOIL_LOAD_AUTO,
        SOIL_CREATE_NEW_ID,
        SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y | SOIL_FLAG_NTSC_SAFE_RGB | SOIL_FLAG_COMPRESS_TO_DXT
      );

	  
	  // OpenGL, texture stuff, etc. should be factored out
	  if( 0 == tex_2d )
	  {
		  printf( "SOIL loading error: '%s'\n", SOIL_last_result() );
	  } else {
		  glEnable( GL_TEXTURE_2D );
		  glBindTexture( GL_TEXTURE_2D, tex_2d );
	  }


      i->second.fullpath = path;
      i->second.id = tex_2d;
    }
  }
}

void MeshDataLoader::AddVertex(const vector<string>& vertex) {
  Vector3d aVector;

  from_string<Scalar>(aVector.x, vertex[0], std::dec);
  from_string<Scalar>(aVector.y, vertex[1], std::dec);
  from_string<Scalar>(aVector.z, vertex[2], std::dec);

  mesh->AddVertex(aVector);
};

void MeshDataLoader::AddNormal(const vector<string>& normal) {
  Vector3d aVector;
    
  from_string<Scalar>(aVector.x, normal[0], std::dec);
  from_string<Scalar>(aVector.y, normal[1], std::dec);
  from_string<Scalar>(aVector.z, normal[2], std::dec);

  mesh->AddNormal(aVector);
};

void MeshDataLoader::AddTexture(const vector<string>& texture){
  // texture vertex
  TexCoord tc;

  from_string<Scalar>(tc.u, texture[0], std::dec);
  from_string<Scalar>(tc.v, texture[1], std::dec);

  mesh->AddTexture(tc);
};

void MeshDataLoader::AddVertexIndices(const vector<string>& vertexIndices) {
// insert vertex indices to last face
  for(size_t i=0; i<vertexIndices.size(); i++) {
    int idx;
    from_string<int>(idx, vertexIndices[i], std::dec);
    
    mesh->AddVertexIndex(idx < 0 ? idx : idx-1/*, currentFace*/);
  }
};

void MeshDataLoader::AddNormalIndices(const vector<string>& normalIndices) {
// insert normal indices to last face
  for(size_t i=0; i<normalIndices.size(); i++) {
    int idx = val_from_string<int>(normalIndices[i], std::dec);

    mesh->AddNormalIndex(idx < 0 ? idx : idx-1/*, currentFace*/);
  }
};

void MeshDataLoader::AddTextureIndices(const vector<string>& textureIndices) {
// insert texture indices to last face
  for(size_t i=0; i<textureIndices.size(); i++) {
    int idx;
    from_string<int>(idx, textureIndices[i], std::dec);

    mesh->AddTextureIndex(idx < 0 ? idx : idx-1/*, currentFace*/);
  }
};

void MeshDataLoader::AddFaceV(const vector<string>& indices) {
  Face face;
  int idx;
  size_t curr = mesh->NumVertices();

  // must have a least three indices
  from_string<int>(idx, indices[0], std::dec);
  face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);

  from_string<int>(idx, indices[1], std::dec);
  face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);

  from_string<int>(idx, indices[2], std::dec);
  face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);

  for(size_t i = 3; i<indices.size();++i) {
    from_string<int>(idx, indices[i], std::dec);
    face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);
  }
  face.numVertices = face.vertexIdx.size();
  mesh->AddFace(face, currentGroups);
};

void MeshDataLoader::AddFaceVT(const vector<string>& indices) {
  Face face;
  int idx;
  size_t curr = mesh->NumVertices();

  // must have a least three indices
  from_string<int>(idx, indices[0], std::dec);
  face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);
  from_string<int>(idx, indices[1], std::dec);
  face.textureIdx.push_back(idx < 0 ? curr+idx : idx-1);

  from_string<int>(idx, indices[2], std::dec);
  face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);
  from_string<int>(idx, indices[3], std::dec);
  face.textureIdx.push_back(idx < 0 ? curr+idx : idx-1);

  from_string<int>(idx, indices[4], std::dec);
  face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);
  from_string<int>(idx, indices[5], std::dec);
  face.textureIdx.push_back(idx < 0 ? curr+idx : idx-1);

  for(size_t i = 6; i<indices.size();++i) {
    from_string<int>(idx, indices[i], std::dec);
    face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);
    from_string<int>(idx, indices[++i], std::dec);
    face.textureIdx.push_back(idx < 0 ? curr+idx : idx-1);
  }
  face.numVertices = face.vertexIdx.size();
  mesh->AddFace(face, currentGroups);
};

void MeshDataLoader::AddFaceVN(const vector<string>& indices) {
  Face face;
  int idx;
  size_t curr = mesh->NumVertices();

  // must have a least three indices
  from_string<int>(idx, indices[0], std::dec);
  face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);
  from_string<int>(idx, indices[1], std::dec);
  face.normalIdx.push_back(idx < 0 ? curr+idx : idx-1);

  from_string<int>(idx, indices[2], std::dec);
  face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);
  from_string<int>(idx, indices[3], std::dec);
  face.normalIdx.push_back(idx < 0 ? curr+idx : idx-1);

  from_string<int>(idx, indices[4], std::dec);
  face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);
  from_string<int>(idx, indices[5], std::dec);
  face.normalIdx.push_back(idx < 0 ? curr+idx : idx-1);

  for(size_t i = 6; i<indices.size();++i) {
    from_string<int>(idx, indices[i], std::dec);
    face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);
    from_string<int>(idx, indices[++i], std::dec);
    face.normalIdx.push_back(idx < 0 ? curr+idx : idx-1);
  }
  face.numVertices = face.vertexIdx.size();
  mesh->AddFace(face, currentGroups);
};

void MeshDataLoader::AddFaceVTN(const vector<string>& indices) {
  Face face;
  int idx;
  size_t curr = mesh->NumVertices();

  // must have a least three indices
  from_string<int>(idx, indices[0], std::dec);
  face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);
  from_string<int>(idx, indices[1], std::dec);
  face.textureIdx.push_back(idx < 0 ? curr+idx : idx-1);
  from_string<int>(idx, indices[2], std::dec);
  face.smoothNormalIdx.push_back(idx < 0 ? curr+idx : idx-1);

  from_string<int>(idx, indices[3], std::dec);
  face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);
  from_string<int>(idx, indices[4], std::dec);
  face.textureIdx.push_back(idx < 0 ? curr+idx : idx-1);
  from_string<int>(idx, indices[5], std::dec);
  face.smoothNormalIdx.push_back(idx < 0 ? curr+idx : idx-1);

  from_string<int>(idx, indices[6], std::dec);
  face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);
  from_string<int>(idx, indices[7], std::dec);
  face.textureIdx.push_back(idx < 0 ? curr+idx : idx-1);
  from_string<int>(idx, indices[8], std::dec);
  face.smoothNormalIdx.push_back(idx < 0 ? curr+idx : idx-1);

  for(size_t i = 9; i<indices.size();++i) {
    from_string<int>(idx, indices[i], std::dec);
    face.vertexIdx.push_back(idx < 0 ? curr+idx : idx-1);
    from_string<int>(idx, indices[++i], std::dec);
    face.textureIdx.push_back(idx < 0 ? curr+idx : idx-1);
    from_string<int>(idx, indices[++i], std::dec);
    face.smoothNormalIdx.push_back(idx < 0 ? curr+idx : idx-1);
  }
  face.numVertices = face.vertexIdx.size();
  mesh->AddFace(face, currentGroups);
};

void MeshDataLoader::SetGroups(const vector<string>& groupNames) {
  currentGroups.clear();
  currentGroups = groupNames;
};

void MeshDataLoader::SetMaterial(const vector<string>& material) {
  for(size_t i = 0; i<currentGroups.size();++i) {
    mesh->SetMaterial(currentGroups, material[0]) ;
  }
};

void MeshDataLoader::AddMaterial(const vector<string>& newMaterial) {
  Material material;
  for(size_t i = 1; i<newMaterial.size();++i) {
    if(newMaterial[i]=="Kd") {
      material.diffuse.r = val_from_string<float>(newMaterial[++i], std::dec);
      material.diffuse.g = val_from_string<float>(newMaterial[++i], std::dec);
      material.diffuse.b = val_from_string<float>(newMaterial[++i], std::dec);
    } else if(newMaterial[i]=="Ka") {
      material.ambient.r = val_from_string<float>(newMaterial[++i], std::dec);
      material.ambient.g = val_from_string<float>(newMaterial[++i], std::dec);
      material.ambient.b = val_from_string<float>(newMaterial[++i], std::dec);
    } else if(newMaterial[i]=="Ks") {
      material.specular.r = val_from_string<float>(newMaterial[++i], std::dec);
      material.specular.g = val_from_string<float>(newMaterial[++i], std::dec);
      material.specular.b = val_from_string<float>(newMaterial[++i], std::dec);
    } else if(newMaterial[i]=="Ns") {
      material.shininess = val_from_string<float>(newMaterial[++i], std::dec);
    } else if(newMaterial[i]=="illum") {
      material.emission = val_from_string<float>(newMaterial[++i], std::dec);
    } else if(newMaterial[i].compare(0,4,"map_") == 0) {

      string pathname("");//(newMaterial[i]);
      string name(newMaterial[++i]);
      mesh->AddTextureName(name, pathname);

      if(newMaterial[i][6] == 'a')
        material.map_ka = val_from_string<float>(name, std::dec);
      else
        material.map_kd = val_from_string<float>(name, std::dec);
    }
  }

  mesh->AddMaterial(newMaterial[0], material);
};
