//	Vector3d.h

//	This class defines a three-dimensional vector, with appropriate
//	operators.  (* is cross product.)

#ifndef VECTOR_H
#define VECTOR_H

#ifndef real
#define real double
#endif

#ifndef Scalar
#define Scalar double
#endif

#include <iostream>
#include <math.h>

template <class T>
bool from_string(T& t, 
                 const std::string& s, 
                 std::ios_base& (*f)(std::ios_base&))
{
  std::istringstream iss(s);
  return !(iss >> f >> t).fail();
}

template <class T>
T val_from_string(const std::string& s, 
                 std::ios_base& (*f)(std::ios_base&))
{
  std::istringstream iss(s);
  T t;
  iss >> f >> t;
  return t;
}

template <class T> T sgn(T a) {
  if (a<0) return -1;
  if (a>0) return 1;
  return 0;
}

struct Vector3d {
  union {
    struct {real x,y,z;};// vector components
    real arr_xyz[3];
  };
	
	Vector3d() { x=y=z=0; }			// constructor
	Vector3d(const real pX, const real pY, const real pZ)
	: x(pX), y(pY), z(pZ) {}

	inline const real& Vector3d::operator[](int i) const
	{ return arr_xyz[i]; }

	inline real& Vector3d::operator[](int i)
	{	return arr_xyz[i]; }	
	
	// vector arithmetic

	Vector3d operator-() 
	{ x-=x, y-=y, z-=z ; return *this; }

	Vector3d operator+(const Vector3d pV) const
	{ Vector3d out( x+pV.x, y+pV.y, z+pV.z ); return out; }

	Vector3d operator-(const Vector3d pV) const
	{ Vector3d out( x-pV.x, y-pV.y, z-pV.z ); return out; }

	Vector3d& operator+=(const Vector3d pV)
	{ x+=pV.x; y+=pV.y; z+=pV.z; return *this; }

	Vector3d& operator-=(const Vector3d pV)
	{ x-=pV.x; y-=pV.y; z-=pV.z; return *this; }

	Vector3d operator/(const real r) const
	{ Vector3d out( x/r, y/r, z/r ); return out; }

	Vector3d& operator/=(const real r)
	{ x/=r; y/=r; z/=r; return *this; }

	real DotProduct(const Vector3d pV) const			// dot product
	{ return x*pV.x + y*pV.y + z*pV.z; }

	Vector3d operator*(const Vector3d pV) const		// cross product
	{ Vector3d out( y*pV.z-z*pV.y, z*pV.x-x*pV.z, x*pV.y-y*pV.x ); 
	  return out; }

	Vector3d CrossProduct( const Vector3d pV ) const
	{ return *this * pV; }

	friend real DotProduct( const Vector3d p1, const Vector3d p2 )
	{ return p1.DotProduct(p2); }

	friend Vector3d CrossProduct( const Vector3d p1, const Vector3d p2 )
	{ return p1 * p2; }

	Vector3d operator*(const real pR) const		// * a scalar
	{ Vector3d out( x*pR, y*pR, z*pR ); return out; }

	friend Vector3d operator*(const real pR, const Vector3d pV)
	{ Vector3d out( pV.x*pR, pV.y*pR, pV.z*pR ); return out; }

	Vector3d& operator*=(const real pR)
	{ x*=pR; y*=pR; z*=pR; return *this; }

	// magnitude

	real SquareMagnitude(void) const
	{ return x*x+y*y+z*z; }

	real Magnitude(void) const
	{ return sqrt(SquareMagnitude()); }

	real Normalize(void) const
	{ return sqrt(SquareMagnitude()/Magnitude()); }

	// projection (component of *this parallel to pV)
	// Note: component perpendicular to pV is:  *this - Proj(pV)

	Vector3d Proj(const Vector3d pV)
	{ Vector3d out( pV * (this->DotProduct(pV) / pV.SquareMagnitude()) ); return out; }

	// cosine of the angle between two vectors:
	real CosAng( const Vector3d pV)
	{ return this->DotProduct(pV) / (Magnitude() * pV.Magnitude()); }

	// comparison
	int operator==( const Vector3d pV) const
	{ return x==pV.x && y==pV.y && z==pV.z; }

	int operator!=( const Vector3d pV) const
	{ return x!=pV.x || y!=pV.y || z!=pV.z; }
};

// input/output
inline std::ostream& operator << (std::ostream& pStr, const Vector3d& pV) {
	return (pStr << "(" << pV.x << ", " << pV.y << ", " << pV.z << ")");
};

inline std::istream& operator>>(std::istream pStr, Vector3d& pV) { 
  return (pStr >> pV.x >> pV.y >> pV.z); 
}

// other
inline real SquareMagnitude (const Vector3d& v)
{ return v.x*v.x + v.y*v.y + v.z*v.z; }

inline real Magnitude (const Vector3d& v)
{ return (real) sqrt(SquareMagnitude(v)); }

inline Vector3d Normalize (const Vector3d& v)
{ return v / Magnitude(v); }

//const double PI = 3.141592654f;
const double PI = 3.14159265358979323846;

#endif
